/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

/**
* Klasse, die eine Phase verwaltet
*/
function Phase(
    parent,     // Übergeordnetes Objekt
    data,       // JSON mit Informationen über diese Phase und ggf. Unterphasen
    objDone     // Optional: Informationen darüber, ob diese Phase bzw. Unterphasen erledigt sind
) {
    var self = this;
    if ((typeof objDone == 'string') || (objDone instanceof String))
        objDone = Phase._listDone2Object(objDone);
    // Werte aus JSON importieren
    Phase._import(this, data);
    // Elternelement
    this.parent = parent;
    // Phase abgeschlossen?
    if ((objDone instanceof Object) && objDone.done)
        this.done = true;
    // Unterphasen
    this.phases = (data.phases instanceof Array) ? data.phases.map(function(p, index) {
        return new Phase(
            self,
            p,
            ((objDone instanceof Object) && objDone.phases && objDone.phases[index]) ? objDone.phases[index] : null
        );
    }) : [];
}

// Aus JSON zu importierende Werte inklusive Typ und Standardwerte
Phase.IMPORT = [
    { "key": "title", "default": "Auftrag", "type": String },
    { "key": "label", "default": "Weiter", "type": String },
    { "key": "content", "default": "", "type": String },
    { "key": "score", "default": 0, "type": Number },
    { "key": "mintime", "default": 0, "type": Number },
    { "key": "timeoffset", "default": 0, "type": Number },
    { "key": "tooearly", "default": "So schnell kannst du das doch nicht abgeschlossen haben!", "type": String },
    { "key": "noreset", "default": false, "type": Boolean },
    { "key": "exit", "default": false, "type": Boolean },
    { "key": "mayskip", "default": false, "type": Boolean }
];

/**
* Den typeof-String zum Primitiv type bestimmen
* @return String    typeof-String bei Primitiven oder ""
*/
Phase._toTypeofString = function(
    type        // Klasse
) {
    switch(type) {
        case String:
            return "string";
        case Number:
            return "number";
        case Boolean:
            return "boolean";
        case Function:
            return 'function';
        case Object:
            return 'object';
    }
    return "";
}

/**
* Werte aus data entsprechend Phase.IMPORT nach self importieren
* @return undefined
*/
Phase._import = function(
    self,   // Phase-Instanz, in die Werte importiert werden sollen
    data    // JSON mit Informationen über diese Phase und ggf. Unterphasen
) {
    if (!(data instanceof Object))
        data = {};
    Phase.IMPORT.map(function(o) {
        self[o.key] = ((data[o.key] instanceof o.type) || (typeof data[o.key] == Phase._toTypeofString(o.type))) ? data[o.key] : o["default"];
    });
}

// Objekt, dem Start einer neuen Phase gemeldet werden soll
Phase.prototype.parent = null;

// Unterphasen
Phase.prototype.phases = [];

// Wurde diese Phase abgeschlossen?
Phase.prototype.done = false;

/**
* Aufgrund von Phasenabschlüssen erreichten Punktewert bestimmen
* @return Number    Erreichter Punktewert
*/
Phase.prototype.countScore = function() {
    return this.phases.reduce(function(score, p) {
        return score + p.countScore();
    }, (this.done && Number.isFinite(this.score)) ? this.score : 0);
}

/**
* Ermittelt, ob eine Phase abgeschlossen wurde, die den Arbeitsauftrag beendet
* @return Boolean   Arbeitsauftrag beendet?
*/
Phase.prototype.hasExited = function() {
    if (this.exit && this.done)
        return true;
    for (var z = 0; z < this.phases.length; ++z) {
        if (this.phases[z].hasExited())
            return true;
    }
    return false;
}

/**
* Eine gut speicherbare Auflistung erstellen, welche Phasen bereits erledigt sind
* @return String    Auflistung erledigter Phasen
*/
Phase.prototype.listDone = function() {
    var sub = this.phases.reduce(function(s, p) { return s + p.listDone(); }, '').replace(/0+$/g, '');
    return (this.done ? '1' : '0') + (sub.length ? ('[' + sub + ']') : '');
}

/**
* Aus der gut speicherbaren Auflistung der bereits erledigten Phasen ein Objekt erstellen, das vom Konstruktor verwendet werden kann
* @return Object    Von Konstruktor verwendbares Objekt mit Informationen über bereits erledigte Phasen
*/
Phase._listDone2Object = function(
    s       // Auflistung erledigter Phasen
) {
    var root = { "done": false, "phases": [] };
    var chain = [];
    var last;
    for (var n = 0; n < s.length; ++n) {
        switch (s.charAt(n)) {
            case '[':
                if (chain.length) {
                    last = chain[chain.length - 1];
                    if (!last.length)
                        last.push({ "done": false, "phases": [] });
                    chain.push(last[last.length - 1].phases);
                } else
                    chain.push(root.phases);
                break;
            case ']':
                chain.pop();
                break;
            case '0':
            case '1':
                if (chain.length)
                    chain[chain.length - 1].push({ "done": s.charAt(n) == '1', "phases": [] });
                else
                    root.done = (s.charAt(n) == '1');
        }
    }
    return root;
}

/**
* Übergeordnetes Objekt informieren, dass eine neue Phase gestartet wurde
* @return undefined
*/
Phase.prototype.registerPhaseStarted = function(
    noreset,        // Soll das Speichern der neuen Startzeit unterdrückt werden?
    timeoffset      // Um wie viele Sekunden soll die Startzeit korrigiert werden?
) {
    if (this.parent && this.parent.registerPhaseStarted && (typeof this.parent.registerPhaseStarted == 'function'))
        this.parent.registerPhaseStarted(noreset, timeoffset);
}

/**
* Ermittelt, wie viele Sekunden der Start der letzten Phase zurückliegt
* @return Number    Anzahl an Sekunden (bzw. -1 im browse- und Infinity im review-Modus)
*/
Phase.prototype.getTime = function() {
    if (this.parent && this.parent.getTime && (typeof this.parent.getTime == 'function'))
        return this.parent.getTime();
    return Infinity;
}

/**
* Diese Phase starten (oder aber eine Fehlermeldung anzeigen)
* @return undefined
*/
Phase.prototype.requestStart = function() {
    var time = this.getTime()
    if (!this.mintime || (this.mintime <= time))
        this.start();
    else if (time < 0)
        alert('Im Vorschaumodus ist dies nicht möglich');
    else
        alert(this.tooearly);
}

/**
* Aktuelle Phase ohne irgendeine Überprüfung starten
* @return Boolean   Würde die Phase erstmalig gestartet? (false, wenn Phase schon gestartet war)
*/
Phase.prototype.start = function() {
    if (this.done)
        return false;
    this.done = true;
    this.registerPhaseStarted(this.noreset);
    if (this.parent && this.parent.createSubSection && (typeof this.parent.createSubSection == 'function'))
        this.render(this.parent.createSubSection());
    return true;
}

/**
* Eine Schaltfläche erstellen, die diese Phase startet (oder eine Fehlermeldung anzeigt)
* @return undefined
*/
Phase.prototype.createStartBtn = function(
    parentNode      // Gewünschtes Elternelement der Schaltfläche
) {
    var self = this;
    ScoTools.createHtml('button', parentNode, null, this.label).addEventListener('click', function(ev) {
        ev.preventDefault();
        ev.stopPropagation();
        self.requestStart();
    });
}

/**
* Die Auflistung möglicher nächster Unterphasen neu aufbauen
* @return undefined
*/
Phase.prototype._recreateMenu = function() {
    var ul = ScoTools.createHtml('ul');
    for (var n = 0; n < this.phases.length; ++n) {
        if (this.phases[n].done)
            continue;
        this.phases[n].createStartBtn(ScoTools.createHtml('li', ul));
        if (!this.phases[n].mayskip)
            break;
    }
    if (ul.firstChild) {
        if (this._nodeMenu.firstChild)
            this._nodeMenu.replaceChild(ul, this._nodeMenu.firstChild);
        else
            this._nodeMenu.appendChild(ul);
    } else if (this._nodeMenu.firstChild)
        this._nodeMenu.innerHTML = '';
}

/**
* Für eine Unterphase einen Platz bereitstellen, in dem sie dargestellt werden kann
* @return HTMLElement   Element, in dem Unterphase sich rendern soll
*/
Phase.prototype.createSubSection = function() {
    if (!this._nodeMenu)
        return null;
    this._recreateMenu();
    return ScoTools.createHtml('section', null, null, undefined, this._nodeMenu);
}

/**
* Diese Phase (und ggf. abgeschlossene Unterphasen) auf dem Bildschirm darstellen
* @return undefined
*/
Phase.prototype.render = function(
    parentNode      // HTML-Element, als dessen Kindknoten diese Phase dargestellt werden soll
) {
    if (!parentNode)
        return;
    if (this.parent instanceof Phase)
        ScoTools.createHtml('h2', parentNode, null, this.title);
    else {
        document.title = this.title;
        ScoTools.createHtml('h1', parentNode, null, this.title);
    }
    ScoTools.createHtml('div', parentNode).innerHTML = this.content;
    this._nodeMenu = ScoTools.createHtml('nav', parentNode, { "class": "nextphases" });
    var self = this;
    this.phases.map(function(p) {
        if (p.done)
            p.render(ScoTools.createHtml('section', parentNode, null, undefined, self._nodeMenu));
    });
    this._recreateMenu();
}
